local M = {}

----------------------------------------------------------------------------------------------------
-- Imports
----------------------------------------------------------------------------------------------------

local map            = require("map")
local exploration    = require("campaign/exploration")

local path           = require("scenario/pseudoRandomTour/path")
local levels         = require("scenario/pseudoRandomTour/levels")
local vehicle        = require("scenario/pseudoRandomTour/vehicle")
local waypoints      = require("scenario/pseudoRandomTour/waypoints")
local achievements   = require("scenario/pseudoRandomTour/achievements")
local levelSettings  = require("scenario/pseudoRandomTour/levelSettings")
local configurations = require("scenario/pseudoRandomTour/configurations")

----------------------------------------------------------------------------------------------------
-- Variables
----------------------------------------------------------------------------------------------------

local lastQuery = 0

----------------------------------------------------------------------------------------------------

local function defaultTargetIndicator()
    exploration.state = {
        closestRoad = {
            player = {first=nil, second=nil, best=nil, pos=nil},
            target = {first=nil, second=nil, best=nil, pos=nil}
        },
        focusExplorationPOI = false,
        focusExplorationScenario = false
    }
end

local function reset()
    --[[
    Resets to the initial state
    ]]--
    waypoints.resetWaypoints()
    defaultTargetIndicator()

    guihooks.trigger("prt_reset")
end

local function apply(routeGenerated, firstNode, alwaysSunny, addTraffic)
    --[[
    Applies and generates the scenario based on the route generated
    Parameters:
        routeGenerated - the route generated
        firstNode      - the first node's data for the path
        alwaysSunny    - whether the time of day is always sunny or not
        addTraffic     - whether to generate traffic or not
    ]]--
    local firstNodePos = firstNode.pos
    local scenario = scenario_scenarios.getScenario()

    -- No branching paths so the configs are the same
    scenario.BranchLapConfig = scenario.lapConfig
    scenario_waypoints.initialise()

    levelSettings.setupRandom(alwaysSunny)

    path.setRoute(routeGenerated)
    --graphpath = gp

    if addTraffic then
        -- Generate the traffic
        gameplay_traffic.setupTraffic(nil, false, true, true)
    else
        gameplay_traffic.deleteVehicles()
    end

    guihooks.trigger("prt_reset")

    guihooks.trigger("prt_show_waypoint", {
        type = waypoints.getLastWaypointData(firstNode.name) and "last" or "next",
        wpName = firstNode.name,
        scale = firstNode.scale,
        x = firstNodePos.x,
        y = firstNodePos.y
    })
end

local function jsBoolToLua(jsBool)
    return jsBool or jsBool == 1
end

local function initialise(routeLength, waypointSpacing, alwaysSunny, addTraffic)
    --[[
    Setups the scenario for the player
    Parameters:
        routeLength     - the target length of the route
        waypointSpacing - the spacing of the waypoints (default, sparse, sparser)
        alwaysSunny     - whether the time of day is always sunny or not
        addTraffic      - whether to allow traffic or not
    Returns:
        <string> - the status (ERR, WARN or OK)
        <string> - (only if status = ERR or WARN) - the description of what went wrong/unexpected
        <number> - the length of the route generated (m)
        <number> - the number of waypoints created
    ]]--
    -- Compatability with JavaScript
    alwaysSunny = jsBoolToLua(alwaysSunny)
    addTraffic = jsBoolToLua(addTraffic)

    local scenario = scenario_scenarios.getScenario()

    -- Generate the route
    local gp = map.getGraphpath()
    local routeGenerated, circuit = path.generatePath(gp, routeLength)

    local avgDriveability = path.getRouteDriveability(gp, routeGenerated)

    if not routeGenerated then
        return {"ERR", "No suitable path found for this length!"}
    end

    -- Only keep the necessary nodes
    routeGenerated = path.trimRouteNodes(routeGenerated, gp, waypoints.getWaypointSpacingValue(waypointSpacing))
    -- Ad in the scalings of each node for the waypoints
    routeGenerated = path.addNodeScalings(routeGenerated)

    -- Generate the waypoints from the nodes
    waypoints.createWaypointsFromNodes(scenario, routeGenerated)

    -- Place the player on the first node, pointing at the second
    vehicle.placePlayer(avgDriveability, routeGenerated[1].pos, routeGenerated[2].pos)
    -- Set the scenario up for the route
    apply(routeGenerated, waypoints.getWaypointData("node_1"), alwaysSunny, addTraffic)

    -- Warn about circuits
    if circuit then
        return {
            "WARN",
            "The target length is too long for this map, so a circuit has been generated. Sudden U-turns may be present in the path",
            waypoints.getLengthOfCurrentRoute(),
            #waypoints.getWaypointsInScenario()
        }
    end

    return {"OK", waypoints.getLengthOfCurrentRoute(), #waypoints.getWaypointsInScenario()}
end

local function start()
    --[[
    Start the scenario
    ]]--
    scenario_scenarios.changeState("running")

    -- Show the road indicator
    local player = scenetree.findObject("scenario_player0")
    exploration.queryRoadNodeToPosition(player, player:getPosition(), "player")
    exploration.queryRoadNodeToPosition(player, waypoints.getWaypointData("node_1").pos, "target")
end

local function save(filename, routeLength)
    --[[
    Saves a route to file
    Parameters:
        filename    - the name of the file to save the route as
        routeLength - the length of the route
    ]]--
    configurations.save(levels.getMapName(), filename, routeLength, waypoints.getWaypointsInScenario())
    return {"OK"}
end

local function load(filename)
    --[[
    Load a route from file
    Parameters:
        filename - the name of the file
    Returns:
        <table> - if an error occurred, returns status "ERR" and the error description
                  else, returns status "OK", the route length and the number waypoints loaded
    ]]--
    local fileContents = configurations.load(levels.getMapName(), filename)

    -- Can't find the route, respond with error
    if not fileContents then
        return {"ERR", "This file does not exist"}
    end

    local wps = fileContents.waypoints
    local firstWp = wps[1]

    waypoints.createWaypointsFromNodes(scenario_scenarios.getScenario(), wps)
    vehicle.placePlayer(vec3(firstWp.pos), vec3(wps[2].pos))
    apply(wps, firstWp)

    path.setRoute(fileContents.route)

    return {"OK", waypoints.getLengthOfCurrentRoute(), #waypoints.getWaypointsInScenario()}
end

local function previousRouteExists()
    --[[
    Returns:
        <boolean> - whether a previous route has been generated
    ]]--
    return path.getRoute() ~= nil
end

local function getCurrentRoute()
    --[[
    Returns:
        <table(length, num waypoints)> - the length (m) and the number fo waypoints in the currently
                                         loaded route
    ]]--
    if previousRouteExists() then
        return {waypoints.getLengthOfCurrentRoute(), #waypoints.getWaypointsInScenario()}
    end
end

local function isPlayingScenario()
    --[[
    Returns:
        <boolean> - whether the scenario is being played
    ]]--
    return scenario_scenarios.getScenario().state == "running"
end

----------------------------------------------------------------------------------------------------
-- BeamNG Functions
----------------------------------------------------------------------------------------------------

local function onScenarioFinished()
    achievements.addScenarioToStats(
            waypoints.getLengthOfCurrentRoute()/1000,
            #waypoints.getWaypointsInScenario(),
            vehicle.getVehicleConfig()
    )
end

local function onRaceWaypointReached(data)
    -- Remove waypoint reached and next waypoint (for updating colour)
    guihooks.trigger("prt_reset")

    -- Get the next two waypoints
    local nextWpData = waypoints.getNextWaypointData(data.waypointName)
    local nextNextWpData = nextWpData and waypoints.getNextWaypointData(nextWpData.name)

    -- Set this waypoint as passed
    waypoints.setLastPassedWaypoint(tonumber(data.waypointName:match("%d+")))

    if nextWpData then
        local nextWpPos = nextWpData.pos
        -- Set the next waypoint as the next target
        exploration.queryRoadNodeToPosition(
                scenetree.findObject("scenario_player0"),
                nextWpPos,
                "target"
        )

        -- Add next waypoint to UI
        guihooks.trigger("prt_show_waypoint", {
            type = waypoints.getLastWaypointData(nextWpData.name) and "last" or "next",
            wpName = nextWpData.name,
            scale = nextWpData.scale,
            x = nextWpPos.x,
            y = nextWpPos.y
        })

        -- Show the next next waypoint as greyed out (like in actual scenarios)
        if nextNextWpData then
            local nextNextPos = nextNextWpData.pos
            guihooks.trigger("prt_show_waypoint", {
                type = "nextNext",
                wpName = nextNextWpData.name,
                scale = nextNextWpData.scale,
                x = nextNextPos.x,
                y = nextNextPos.y
            })
        end
    else
        ---- Deactivate target indicator since this is the last waypoint
        --core_groundMarkers.resetAll()
    end
end

local function onPreRender(dt)
    local state = exploration.state

    if not state or not state.closestRoad then
        return
    end


    -- See exploration.onPreRender() --

    lastQuery = lastQuery - dt
    if lastQuery <= 0 then
        local pVehicle = scenetree.findObject("scenario_player0")
        if not pVehicle then return end

        local vehiclePos = pVehicle:getPosition()
        exploration.queryRoadNodeToPosition(pVehicle, vehiclePos, 'player')
        lastQuery = 2.0
    end

    local playerRoadData = state.closestRoad['player']
    if playerRoadData and playerRoadData.first and playerRoadData.first ~= 'nil' then
        for key, data in pairs(state.closestRoad) do
            if key ~= 'player' and data.best then
                core_groundMarkers.setFocus(data.best)
            end
        end
    end
end

local function onRaceStart()
    --[[ The player will trigger 1 or more waypoints, which will be reset on the UI when the scenario
         starts. This forces it to update to the correct value ]]--
    guihooks.trigger("WayPoint", "Checkpoint "..waypoints.getLastPassedWaypoint().." / "..#waypoints.getWaypointsInScenario())
end

local function onScenarioRestarted()
    reset()
end

local function onScenarioUIReady()
    guihooks.trigger('MenuHide')
    guihooks.trigger('ChangeState', 'menu')
end

local function onScenarioChange(sc)
    if sc and sc.state == "pre-running" then
        -- Set state as freeroam to disable automatic scenario controls taking over
        core_gamestate.setGameState('freeroam')
        -- Hide start screen
        guihooks.trigger("MenuHide")
        guihooks.trigger("ChangeState", "menu")

        defaultTargetIndicator()


        --levels.generateForVanillaMaps()
    end
end



local function freeze()
    scenario_scenarios.freezeAll(1)
end

----------------------------------------------------------------------------------------------------
-- Interface
----------------------------------------------------------------------------------------------------

M.initialise = initialise
M.start = start

-- Main UI interface
M.getMapNameFormatted = levels.getMapNameFormatted
M.save = save
M.load = load
M.previousRouteExists = previousRouteExists
M.getCurrentRoute = getCurrentRoute
M.getAchievementStats = achievements.getAchievementStats
M.findNewModdedMaps = levels.findNewModdedMaps
M.generateScenario = levels.copyScenarioFilesTo
M.isPlayingScenario = isPlayingScenario
M.freeze = freeze

-- BeamNG interface
M.onScenarioFinished = onScenarioFinished
M.onRaceStart = onRaceStart
M.onScenarioRestarted = onScenarioRestarted
M.onRaceWaypointReached = onRaceWaypointReached
M.onPreRender = onPreRender
M.onScenarioChange = onScenarioChange
M.onScenarioUIReady = onScenarioUIReady
return M