local M = {}

----------------------------------------------------------------------------------------------------
-- Imports
----------------------------------------------------------------------------------------------------

local tableComp = require("scenario/pseudoRandomTour/tableComp")

----------------------------------------------------------------------------------------------------

local function createSavefile()
    --[[
    Creates a new savefile
    ]]--
    local content = {
        routes = {
            longestRoute = 0,
            totalDistance = 0,
            numberOfRoutes = 0
        },
        waypoints = {
            mostNumberOfWaypoints = 0,
            totalWaypoints = 0
        },
        vehicles = {
            used = {}
        }
    }

    serializeJsonToFile("prt/stats.json", content, true)
end

local function openSaveFile()
    --[[
    Returns the savefile data
    Creates the savefile if it does not yet exist
    Returns:
        <table> - the achievement savefile
    ]]--
    local savefile = jsonDecode(readFile("prt/stats.json"))

    if not savefile then
        createSavefile()
        -- Re-open
        savefile = jsonDecode(readFile("prt/stats.json"))
    end

    return savefile
end

local function biggest(a, b)
    --[[
    Parameters:
        a - number one
        b - number two
    Returns:
        <number> - the larger of the two numbers
    ]]--
    return a > b and a or b
end

local function addScenarioToStats(routeLength, numWaypoints, vehicleUsed)
    --[[
    Adds a completed scenario to the stats
    Parameters:
        routeLength  - the length of the route (km)
        numWaypoints - the number of waypoints in the scenario
        vehicleUsed  - the vehicle used (should be of the format <vehicleName>_<config>)
    ]]--
    local savefile = openSaveFile()

    savefile.vehicles.used[vehicleUsed] = true

    savefile.routes.longestRoute = biggest(savefile.routes.longestRoute, routeLength)
    savefile.routes.totalDistance = savefile.routes.totalDistance + routeLength
    savefile.routes.numberOfRoutes = savefile.routes.numberOfRoutes + 1

    savefile.waypoints.totalWaypoints = savefile.waypoints.totalWaypoints + numWaypoints
    savefile.waypoints.mostNumberOfWaypoints = biggest(savefile.waypoints.mostNumberOfWaypoints, numWaypoints)

    serializeJsonToFile("prt/stats.json", savefile, true)
end

local function getAchievementStats()
    --[[
    Returns:
        <table> - a list of achievements and their current progress
    ]]--
    local data = openSaveFile()

    return {
        singleWps = {
            name = "Single Scenario Waypoints",
            value = data.waypoints.mostNumberOfWaypoints,
            targets = {25, 50, 200, 400}
        },
        singleDist = {
            name = "Single Scenario Distance (km)",
            value = math.floor(data.routes.longestRoute*10)/10,
            targets = {2.5, 5, 20, 40}
        },
        lifetimeWps = {
            name = "Lifetime Waypoints",
            value = data.waypoints.totalWaypoints,
            targets = {50, 100, 400, 800}
        },
        lifetimeDist = {
            name = "Lifetime Distance (km)",
            value = math.floor(data.routes.totalDistance*10)/10,
            targets = {20, 40, 160, 320}
        },
        vehiclesUsed = {
            name = "Unique Vehicles Used",
            value = tableComp.lengthOf(data.vehicles.used),
            description = "Configurations of vehicles are allowed - ie. Bolide 320 and Bolide 350 are two different vehicles",
            targets = {2, 4, 16, 32}
        }
    }
end

M.addScenarioToStats = addScenarioToStats
M.getAchievementStats = getAchievementStats
return M