local M = {}

local function setupCloudCoverage(cloudState)
    --[[
    Setups the cloud layer for the scenario
    Parameters:
        cloudState - the cloud state (clear, cloudy, dense)
                     can be nil which will not create any cloud layer
    ]]--
    print("Setting cloud coverage to: "..cloudState)

    if scenetree.findObject("clouds") then
        TorqueScript.eval("clouds.delete();")
    end

    if not cloudState or cloudState == "clear" then
        -- Default to no clouds
        return
    end

    local cloudColour = ""
    local cloudCoverage = 0

    if cloudState == "cloudy" then
        cloudColour = "0.9 0.9 0.9 1"
        cloudCoverage = 1
    elseif cloudState == "dense" then
        cloudColour = "0.26 0.26 0.26 1"
        cloudCoverage = 3
    end

    -- Create the cloud layer
    TorqueScript.eval([[
        new CloudLayer(clouds) {
            baseColor = "]]..cloudColour  ..[[";
            coverage  = "]]..cloudCoverage..[[";
            texture = "art/skies/clouds/clouds_normal_displacement";
        };
    ]])
end

local function setupWeather(weatherState)
    --[[
    Setups the weather in the scenario
    Parameters:
        weatherState - the weather state (sunny, light-rain, heavy-rain, lightning)
                       can be nil which will set the weather to the default
    ]]--
    print("Setting weather to: "..weatherState)

    if scenetree.findObject("weather") then
        TorqueScript.eval("weather.delete();")
    end

    if not weatherState or weatherState == "sunny" then
        -- Default to standard scenario weather
        setupCloudCoverage("clear")
        return
    end

    local numDrops = 0
    local dropSize = 0
    local datablock = ""

    if weatherState:find("rain") then
        if weatherState == "light-rain" then
            numDrops = 1024
            dropSize = 0.5
            setupCloudCoverage("cloudy")
        elseif weatherState == "heavy-rain" then
            numDrops = 4096
            dropSize = 1
            setupCloudCoverage("dense")
        end

        datablock = "rain_drop"
    elseif weatherState:find("snow") then
        if weatherState == "light-snow" then
            numDrops = 1024
            dropSize = 1
        elseif weatherState == "heavy-snow" then
            numDrops = 4096
            dropSize = 2
        end

        setupCloudCoverage("cloudy")
        datablock = "Snow_menu"
    end

    -- Create the precipitation object
    TorqueScript.eval([[
        new Precipitation(weather) {
            numDrops  = "]]..numDrops ..[[";
            dropSize  = "]]..dropSize ..[[";
            datablock = "]]..datablock..[[";
        };
    ]])
end

local function setRandomWeather()
    --[[
    Setups random weather
    Probabilities:
        sunny - 0.4
        (light/heavy) (rain/snow) - 0.15
    ]]--
    local x = math.random()

    local sunnyChance = 0.4
    local equalChanceOther = (1-sunnyChance)/4

    setupWeather("heavy-rain")

    if x < sunnyChance then
        setupWeather("sunny")
    elseif x < sunnyChance + equalChanceOther then
        setupWeather("light-rain")
    elseif x < sunnyChance + equalChanceOther*2 then
        setupWeather("light-snow")
    elseif x < sunnyChance + equalChanceOther*3 then
        setupWeather("heavy-rain")
    else
        setupWeather("heavy-snow")
    end
end

local function setTOD(time)
    --[[
    Sets a time of day
    Parameters:
        time - the time to set
    ]]--
    local tod = scenetree.findObject("tod")

    if not tod then
        return
    end

    tod.time = time
    tod.dayScale = "0"
    tod.nightScale = "0"
    tod.play = "0"
end

local function setupRandomTOD()
    --[[
    Setups a random time of day
    ]]--
    setTOD(math.random())
end

local function setupRandom(alwaysSunny)
    --[[
    Setups random weather and time of day
    Parameters:
        alwaysSunny - whether the time of day is always sunny
    ]]--
    if alwaysSunny then
        -- Generate in the ranges 0->0.2, 0.8->1
        local x = math.random(0, 400)/1000
        if x > 0.2 then
            x = x+0.6
        end

        setTOD(x)
    else
        setupRandomTOD()
    end

    setRandomWeather()
end

M.setupRandom = setupRandom
return M