local M = {}

local tableComp = require("scenario/pseudoRandomTour/tableComp")

local HOST_SC_FILES_PATHS = {
    "/levels/east_coast_usa/scenarios/pseudoRandomTour.json",
    "/levels/east_coast_usa/pseudoRandomTour.prefab",
    "/levels/east_coast_usa/scenarios/PRT_preview.png"
}

local HOST_SC_PLAYER_POS = "-625.949036 393.17807 21.9553108"

local function movePlayerSpawnpoint(prefabFilePath, pos)
    --[[
    Moves the player spawnpoint in a prefab
    Parameters:
        prefabFilePath - the file path for the prefab level for the PRT mod
        pos            - the position of the default spawnpoint for the level this prefab belongs to
    ]]--
    local file = io.open(prefabFilePath, "r")
    local content = file:read("*all")
    file:close()

    content = content:gsub(HOST_SC_PLAYER_POS, pos.x.." "..pos.y.." "..pos.z)

    file = io.open(prefabFilePath, "w")
    file:write(content)
    file:close()
end

local function loadASpawnpoint(mapName)
    --[[
    Parameters:
        <string> mapName - the name of the map
    Returns:
        <vec3> - the position of a spawnpoint on the map
    ]]--
    local files = FS:findFiles("/levels/"..mapName.."/main/MissionGroup/", "*", 9999, true, false)

    -- Find a spawn sphere and use it
    for _, fileName in ipairs(files) do
        local file = jsonReadFile(fileName)

        if file.class == "SpawnSphere" then
            return vec3(file.position)
        end
    end

    -- Try to find the spawnpoint from the main file
    local mainLevel = jsonReadFile("levels/"..mapName.."/main.level.json")

    if mainLevel then
        local spawnPoint = tableComp.getTableFromEntry(mainLevel, "class", "SpawnSphere")

        if spawnPoint and spawnPoint.position then
            return vec3(spawnPoint.position)
        end
    end

    -- Backwards compatability - .mis
    local mis = FS:findFiles("/levels/"..mapName.."/", "*.mis", 1, true, false)
    if #mis > 0 then
        -- Should only be one
        local misFile = io.open(mis[1], "r")
        local contents = misFile:read("*a")


        -- Find the first occurrence of a spawnsphere
        local _, i = contents:find("new SpawnSphere")
        local spawnSphereData = contents:sub(i)

        -- Find the next occurrence of position
        local _, j = spawnSphereData:find("position")
        local posData = spawnSphereData:sub(j)

        -- Find the position string
        local positionStr = posData:match('(%".-%")'):gsub('"', "")

        -- Convert to numbers and vectorise it
        local pos = {}
        for coord in positionStr:gmatch("%S+") do
            table.insert(pos, tonumber(coord))
        end

        return vec3(pos)
    end


    -- Can't find it, default to somewhere in the sky
    return vec3(0,0,500)
end

local function findModdedMaps()
    --[[
    Finds and returns the names of modded maps
    Returns:
        <table> - a list of modded map names
    ]]--
    local foundMaps = FS:findFiles("/levels/", "*", 0, true, true)
    local moddedMaps = {}

    for _, map in ipairs(foundMaps) do
        map = map:gsub("/levels/", "")
        local mapInfo = jsonReadFile("levels/"..map.."/info.json")

        -- Makes the assumption that modded maps don't include BeamNG as an author ;)
        if mapInfo and mapInfo.authors and not mapInfo.authors:find("BeamNG") then
            table.insert(moddedMaps, {internalName = map, formattedName = mapInfo.title, features = mapInfo.features})
        end
    end

    for i, mapData1 in ipairs(moddedMaps) do
        for j, mapData2 in ipairs(moddedMaps) do
            if i ~= j then
                -- There are multiple versions of this map so append the features to distinguish them
                if mapData1.formattedName == mapData2.formattedName then
                    mapData1.formattedName = mapData1.formattedName.." ("..mapData1.features..")"
                    break
                end
            end
        end
    end

    return moddedMaps
end

local function getMapName()
    --[[
    Returns:
        <string> - the raw map name
    ]]--
    local misName = getMissionFilename():gsub("/levels/", "")

    if misName:find("info.json") then
        return misName:gsub("/info.json", "")
    elseif misName:find("main.level.json") then
        return misName:gsub("/main.level.json", "")
    else
        -- Backwards compatability for .mis files
        return misName:gsub(".mis", ""):match("%a+")
    end
end

local function getMapNameFormatted(mapName)
    --[[
    Parameters:
        mapName - the map to format (can be nil to use the currently loaded map)
    Returns:
        <string> - the formatted map name
    ]]--
    local rawMapName = mapName or getMapName()

    -- See if it is a modded map, since these are more inconsistent and so need formatting differently
    for _, mapData in ipairs(findModdedMaps()) do
        if mapData.internalName:lower() == rawMapName:lower() then
            return mapData.formattedName
        end
    end

    -- Not a modded map, format the name as a string
    local formattedName = ""

    for word in rawMapName:gmatch("([^_]+)") do
        -- Special case, since USA is an acronym
        if word == "usa" then
            word = "USA"
        else
            -- Uppercase first letter
            word = word:sub(1,1):upper()..word:sub(2)
        end

        formattedName = formattedName..word.." "
    end

    -- Remove extra space at the end
    formattedName = formattedName:sub(1, #formattedName-1)

    return formattedName
end
local function copyScenarioFilesTo(mapName)
    --[[
    Copies the scenario files to work with a given map
    Parameters:
        mapName - the name of the map
    ]]--
    for _, filePath in ipairs(HOST_SC_FILES_PATHS) do
        local newFilePath =  filePath:gsub("east_coast_usa", mapName)
        FS:copyFile(filePath, newFilePath)

        if newFilePath:find(".prefab") then
            movePlayerSpawnpoint(newFilePath, loadASpawnpoint(mapName))
        elseif newFilePath:find(".json") then
            local scFile = jsonReadFile(newFilePath)
            scFile[1].name = "Pseudo Random Tour: "..getMapNameFormatted(mapName)
            serializeJsonToFile(newFilePath, scFile, true)
        end
    end
end

local function generateForVanillaMaps()
    --[[
    Generates the scenario files for appropriate vanilla maps
    ]]--
    local maps = {
        "automation_test_track",
        "driver_training",
        "hirochi_raceway",
        "italy",
        "jungle_rock_island",
        "small_island",
        "Utah",
        "west_coast_usa"
    }

    for _, map in ipairs(maps) do
        copyScenarioFilesTo(map)
    end
end

local function findNewModdedMaps()
    --[[
    Returns:
        <table(internalName, formattedName)> - a list of modded maps that have not had a PRT
                                               scenario made for them (both internal and user names)
    ]]--
    local maps = findModdedMaps()
    local newMaps = {}

    for _, map in ipairs(maps) do
        -- If no scenario exists, it is a new map
        if not FS:fileExists("/levels/"..map.internalName.."/scenarios/pseudoRandomTour.json") then
            table.insert(newMaps, map)
        end
    end

    return newMaps
end

M.getMapName = getMapName
M.getMapNameFormatted = getMapNameFormatted
M.copyScenarioFilesTo = copyScenarioFilesTo
M.generateForVanillaMaps = generateForVanillaMaps
M.findNewModdedMaps = findNewModdedMaps
return M