local M = {}

----------------------------------------------------------------------------------------------------
-- Imports
----------------------------------------------------------------------------------------------------

local tableComp = require("scenario/pseudoRandomTour/tableComp")

----------------------------------------------------------------------------------------------------
-- Variables
----------------------------------------------------------------------------------------------------

local scRoute

----------------------------------------------------------------------------------------------------

local function round(x)
    --[[
    Rounds a number
    Parameters:
        x - the number to round
    Returns:
        <number> - the rounded version of x
    ]]--
    local truncatedX = math.floor(x)
    local decimalPart = x - truncatedX

    if decimalPart >= 0.5 then
        return truncatedX + 1
    else
        return truncatedX
    end
end

local function analyseRoute(route)
    --[[
    Tally's up the number of times each node in the route occurs
    Parameters:
        route - the route of nodes
    Returns:
        <table> - the nodes an their frequency
    ]]--
    local nodes = {}

    for _, v in ipairs(route) do
        if not nodes[v] then
            nodes[v] = 0
        end

        nodes[v] = nodes[v] + 1
    end
end

local function hasSmallerCycle(path, cycleLength)
    --[[
    Finds a cycle in a path and returns whether the cycle found is smaller than the allowed length
    Parameters:
        path <table>         - the path to find cycle son
        cycleLength <number> - the minimum cycle length that is NOT allowed
    Returns:
        <boolean> - whether the path has a cycle of smaller than allowed length
    ]]--
    local visitedNodes = {}

    for i, node in ipairs(path) do
        if visitedNodes[node] then
            -- Since the path is a linear graph, minus the difference in index
            local foundLength = i - visitedNodes[node] <= cycleLength

            if foundLength then
                -- Cycle too small
                return true
            end

            -- Continue to find cycles
        else
            -- Stores the index
            visitedNodes[node] = i
        end
    end

    return false
end

local function isValidPath(path)
    --[[
    Parameters:
        path - the path to validate
    Returns:
        <boolean> - whether the path is valid
    ]]--
    -- Minimum cycle length 5
    return not hasSmallerCycle(path, 5)
end

local function getLengthOfRoute(nodes)
    --[[
    Parameters:
        nodes - the list waypoint route to get the length of
    Returns:
        <number> - the length of the route (in m)
    ]]--
    local length = 0
    local lastNode

    for _, nodeData in ipairs(nodes) do
        local currNode = vec3(nodeData.pos)

        -- The first node has no reference point, so ignore
        if lastNode then
            length = length + (currNode-lastNode):length()
        end

        lastNode = currNode
    end

    return math.floor(length)
end

local function loopCircuit(circuit, lapCount)
    --[[
    Loops a circuit until a target length is reached
    Parameters:
        circuit  - the circuit to loop (should contain a position for each node)
        lapCount - the number of laps the loop will have
    Returns:
        <table> - a table of nodes, looped from {circuit}
    ]]--
    local startNode = circuit[1]
    -- Remove end node, since it will be repeated as it is also the start node
    circuit[#circuit] = nil

    local loopedPath = tableComp.repeatTable(circuit, lapCount)
    -- Add the start node back to the end
    loopedPath[#loopedPath+1] = startNode
    return loopedPath
end

local function getNodePositions(graphpath, nodes)
    --[[
    Parameters:
        graphpath - the graphpath for this level
        nodes     - the nodes
    Return:
        <table> - the nodes with their position
    ]]--
    local nodePositions = {}

    for _, nodeName in ipairs(nodes) do
        table.insert(nodePositions, {name = nodeName, pos = graphpath.positions[nodeName]})
    end

    return nodePositions
end

local function generateCircuit(graphpath)
    --[[
    Generates a circuit, going through 3 selected points and linking them together
    3 points reduces the chance of sudden U-turns, though not completely
    Parameters:
        graphpath - the graphpath for this level
    Returns:
        <table> - the circuit path generated
    ]]--
    local n1 = tableComp.getRandomEntry(graphpath.graph)
    local n2 = tableComp.getRandomEntry(graphpath.graph)
    local n3 = tableComp.getRandomEntry(graphpath.graph)

    -- Path of n1->n2->n3->n1
    local pathA = graphpath:getPath(n1, n2)
    local pathB = graphpath:getPath(n2, n3)
    local pathC = graphpath:getPath(n3, n1)

    return tableComp.mergeAppend(pathA, pathB, pathC)
end

local function generatePath(graphpath, maxLength)
    --[[
    Generates a random path for a level, constraining to a max length
    The path starts at a random position, but will always start at a waypoint node
    Parameters:
        graphpath - the graphpath for this level
        maxLength - the maximum length (in m) the path can be
    Returns:
        <table>   - the path of nodes generated (could be nil if no suitable path is found)
        <boolean> - whether the path return was generated as a circuit
    ]]--
    local timeout = 0
    local path

    -- A path may not always be possible from a given waypoint or be valid
    while not path or not isValidPath(path) do
        path = graphpath:getRandomPathG(
            -- Start at a random point
            tableComp.getRandomEntry(graphpath.graph),
            vec3(0,0,0),
            maxLength,
            nil,
            nil,
            false
        )

        timeout = timeout + 1

        -- Failsafe to prevent infinite looping (try 25 times, then give up)
        if timeout >= 25 then
            --[[ Instead, generate a circuit to loop for the desired length ]]--
            local circuit = generateCircuit(graphpath)

            local pathNodes = getNodePositions(graphpath, circuit)
            local lapCount = round(maxLength/getLengthOfRoute(pathNodes))

            path = loopCircuit(circuit, lapCount)

            return path, true
        end
    end

    return path, false
end

local function getNodeDetails(gp, nodeName, nodeNum, totalNodes)
    --[[
    Collates details about a node
    Parameters:
        gp         - the graphpath for this scenario
        nodeName   - the name of the node
        nodeNum    - the node number in its route
        totalNodes - the total nodes in the route this node belongs to
    Returns:
        <table> - the details of the node (name, pos and type [end, intersection, standard])
    ]]--

    --if scenetree.findObject("obj_"..nodeNum) then
    --    TorqueScript.eval("obj_"..nodeNum..".delete();")
    --end


    -- A valency > 2 is indicative of an intersection node
    local atIntersection = tableComp.lengthOf(gp.graph[nodeName]) > 2
    --[[End nodes are the first and last node and also the second node, since this is the actual
        first node the player drives to (player spawns on first node of the route)]]--
    local isAnEndNode = nodeNum == 1 or nodeNum == 2 or nodeNum == totalNodes

    local nodeType
    -- Start/end nodes take precedence, followed by intersection nodes
    if isAnEndNode then
        nodeType = "end"
    elseif atIntersection then
        nodeType = "intersection"

        --TorqueScript.eval([[
        --    new TSStatic(obj_]]..nodeNum..[[) {
        --        shapeName = "art/shapes/interface/checkpoint_marker_sphere.dae";
        --        scale = "3 3 3";
        --        position = "]]..gp.positions[nodeName].x..[[ ]]..gp.positions[nodeName].y..[[ ]]..gp.positions[nodeName].z..[[";
        --        instanceColor = "1 0 0 0.8";
        --    };
        --]])
    else
        nodeType = "standard"

        --TorqueScript.eval([[
        --    new TSStatic(obj_]]..nodeNum..[[) {
        --        shapeName = "art/shapes/interface/checkpoint_marker_sphere.dae";
        --        scale = "3 3 3";
        --        position = "]]..gp.positions[nodeName].x..[[ ]]..gp.positions[nodeName].y..[[ ]]..gp.positions[nodeName].z..[[";
        --        instanceColor = "0 1 0 0.4";
        --    };
        --]])
    end

    return {
        name = nodeName,
        pos = gp.positions[nodeName],
        nodeType = nodeType
    }
end

local function trimRouteNodes(route, gp, targetNodeSpacing)
    --[[
    Trims a route's nodes into those are are either important (1) or are sufficiently far enough
    away from the previous node (2)
    In case (1), importance is defined by a node have a valency > 2, indicating a junction
    In case (2), if a node is {targetNodeSpacing}m away, then it is sufficiently far enough away
    Parameters:
        route             - the route to trim
        gp                - the graphpath for this level
        targetNodeSpacing - the target minimum spacing between two nodes (in m)
    Returns:
        <table> - the trimmed route
    ]]--
    local nodes = {}
    local trimmedNodes = {}

    -- Stores the last node type checked
    local lastNodeTypeChecked

    -- Get details about each node
    for i, nodeName in ipairs(route) do
        nodes[i] = getNodeDetails(gp, nodeName, i, #route)
    end

    for i, currNode in ipairs(nodes) do
        -- First node always get added
        if i ~= 1 then
            -- There will always be a previous node, since the first node skips this section
            local prevNode = trimmedNodes[#trimmedNodes]
            local wasLastNodeAnIntersection = lastNodeTypeChecked == "intersection"

            local dist = (prevNode.pos-currNode.pos):length()

            local isIntersectionNode = currNode.nodeType == "intersection"
            local isEndNode = currNode.nodeType == "end"
            --[[Many intersection nodes can appear next to each other, take the last waypoint in the
                chain - assumes that intersections are spaced far enough apart; ie. there is a
                standard node between two intersections]]--
            local isUniqueIntersectionNode = isIntersectionNode and not wasLastNodeAnIntersection
            local isAdjacentToIntersection = isIntersectionNode and wasLastNodeAnIntersection
            local isInsideTarget = dist >= targetNodeSpacing

            --[[If this node is within 50m of the target spacing, and the two node types
                are different, then this node should NEVER be added]]--
            local isTooClose = dist <= 50 and
                    (isUniqueIntersectionNode) or
                    (not isIntersectionNode and wasLastNodeAnIntersection)

            -- If this node doesn't meet any condition, skip it
            if isTooClose or not (isEndNode or isUniqueIntersectionNode or isInsideTarget) then
                if isAdjacentToIntersection then
                    -- Overwrite previous intersection node
                    trimmedNodes[#trimmedNodes] = currNode
                end

                goto continue
            end
        end

        table.insert(trimmedNodes, currNode)

        ::continue::

        lastNodeTypeChecked = currNode.nodeType
    end

    --for i, data in ipairs(trimmedNodes) do
    --    if scenetree.findObject("obj_node_"..i) then
    --        TorqueScript.eval("obj_"..i..".delete();")
    --    end
    --
    --    TorqueScript.eval([[
    --        new TSStatic(obj_node_]]..i..[[) {
    --            shapeName = "art/shapes/interface/checkpoint_marker_sphere.dae";
    --            scale = "3 3 3";
    --            position = "]]..data.pos.x..[[ ]]..data.pos.y..[[ ]]..(data.pos.z+3)..[[";
    --            instanceColor = "1 1 1 1";
    --        };
    --    ]])
    --end


    return trimmedNodes
end

local function addNodeScalings(nodes)
    --[[
    Adds the scalings for a set of nodes, if the node is not a BeamNGWaypoint, the default scale of
    5 is used, with the minimum specified being 2.5
    Parameters:
        nodes - the list of nodes
    Returns:
        <table> - the list of nodes with their scalings added as a key
    ]]--
    for _, data in ipairs(nodes) do
        local obj = scenetree.findObject(data.name)

        -- Minimum scale of 2.5, default of 5
        data.scale = obj and math.max(2.5, obj:getScale().x) or 5
    end

    return nodes
end

local function getRouteDriveability(gp, route)
    --[[
    Averages the driveability of the route
    Parameters:
        gp    - the graphpath for this level
        route - the route (a list containing)
    Returns:
        <number> - the average driveability
    ]]--
    local totalDriveability = 0

    for i, node in ipairs(route) do
        -- Last node will not have anything to compare to, and is handled by the previous node
        if i < #route then
            if node:find("DecalRoad") then
                node = node:match("%a+%d+")
            end

            local nodeDriveability = 0
            local thisNode = gp.graph[node]

            -- Not all nodes are in the graph
            if thisNode then
                local nextNode = route[i+1]
                local edge = thisNode[nextNode]

                -- An edge may not always exist
                if edge then
                    -- The driveability is the 3rd entry, but may not always be present
                    nodeDriveability = edge[3] or 0
                end
            end

            totalDriveability = totalDriveability + nodeDriveability
        end
    end

    return totalDriveability/#route
end

local function setRoute(newRoute)
    scRoute = newRoute
end

local function getRoute()
    return scRoute
end

----------------------------------------------------------------------------------------------------
-- Interface
----------------------------------------------------------------------------------------------------

M.getLengthOfRoute = getLengthOfRoute
M.generatePath = generatePath
M.trimRouteNodes = trimRouteNodes
M.addNodeScalings = addNodeScalings
M.analyseRoute = analyseRoute
M.getRouteDriveability = getRouteDriveability
M.setRoute = setRoute
M.getRoute = getRoute
return M