local M = {}

local path = require("scenario/pseudoRandomTour/path")

----------------------------------------------------------------------------------------------------
-- Variables
----------------------------------------------------------------------------------------------------

local wpsInScenario = {}
local waypointSpacing = 0
local lastPassedWaypoint = ""

----------------------------------------------------------------------------------------------------

local function resetWaypoints()
    --[[
    Delete all waypoints previously created
    Also empties the lap and branch configs in the scenario
    ]]--
    local i = 1
    local name = "node_"..i

    while scenetree.findObject(name) do
        TorqueScript.eval(name..".delete();")

        i = i + 1
        name = "node_"..i
    end

    -- Reset configs
    local scenario = scenario_scenarios.getScenario()
    scenario.lapConfig = {}
    scenario.BranchLapConfig = {}

    wpsInScenario = {}
end

local function createWaypoint(scenario, waypointName, position, scale)
    --[[
    Creates a waypoint at a given position
    Parameters:
        scenario     - the scenario instance
        waypointName - the name of the waypoint
        position     - the position of the waypoint
        scale        - the scale of the waypoint
    ]]--
    -- Delete for re-creation
    if scenetree.findObject(waypointName) then
        TorqueScript.eval(waypointName..".delete();")
    end

    TorqueScript.eval([[
        new BeamNGWaypoint(]]..waypointName..[[) {
          position = "]]..position.x..[[ ]]..position.y..[[ ]]..position.z..[[";
          scale = "]]..scale..[[ ]]..scale..[[ ]]..scale..[[";
        };

        ScenarioObjectsGroup.add(]]..waypointName..[[);
    ]])

    -- Set the node data for the waypoint
    scenario.nodes[waypointName] = {
        pos = position,
        radius = scale
    }

    -- Add it to the lap-config for it to be registered
    table.insert(scenario.lapConfig, waypointName)

    table.insert(wpsInScenario, {
        name = waypointName,
        pos = {
            x = position.x,
            y = position.y,
            z = position.z
        },
        scale = scale
    })
end

local function createWaypointsFromNodes(scenario, nodes)
    --[[
    Creates waypoints from a list of nodes (named node_[1->#nodes])
    Parameters:
        scenario - the scenario instance
        nodes    - the list of node (should contain a position vector and a scale value)
    ]]--
    resetWaypoints()

    for i, data in ipairs(nodes) do
        createWaypoint(scenario, "node_"..i, vec3(data.pos), data.scale)
    end
end

local function findWaypointsInLevel()
    --[[
    Returns:
        <table> - the list of waypoint names in the scenario
    ]]--
    return scenetree.findClassObjects("BeamNGWaypoint")
end

local function getNextWaypointData(currentName)
    --[[
    Parameters:
        currentName - the current waypoint name
    Returns:
        <table> - the next waypoint's data (nil if there is no next waypoint)
    ]]--
    for i, data in ipairs(wpsInScenario) do
        if data.name == currentName and i < #wpsInScenario then
            return wpsInScenario[i+1]
        end
    end
end

local function getLastWaypointData(currWaypointName)
    --[[
    Parameters:
        currWaypointName - the waypoint to check
    Returns:
        <boolean> - whether this waypoint is the last waypoint
    ]]--
    return wpsInScenario[#wpsInScenario].name == currWaypointName
end

local function getWaypointData(waypointName)
    --[[
    Parameters:
        waypointName - the name of the waypoint
    Returns:
        <table> - the waypoint data
    ]]--
    for _, data in ipairs(wpsInScenario) do
        if data.name == waypointName then
            return data
        end
    end

    print("No data exists for waypoint ", waypointName)
end

local function setLastPassedWaypoint(waypointNum)
    --[[
    Parameters:
        waypointNum - the waypoint number
    ]]--
    lastPassedWaypoint = waypointNum
end

local function getWaypointSpacingValue(wpSpacing)
    --[[
    Parameters:
        wpSpacing - the type of spacing to use (default, sparse, sparser)
    Returns:
        <number> - the waypoint spacing (in m)
    ]]--
    if wpSpacing == "sparse" then
        return 250
    elseif wpSpacing == "sparser" then
        return 500
    else
        return 100
    end
end

local function getLastPassedWaypoint()
    return lastPassedWaypoint
end

local function getLengthOfCurrentRoute()
    --[[
    Returns:
        <number> - the length of the currently loaded route (in m)
    ]]--
    return path.getLengthOfRoute(wpsInScenario)
end

local function getWaypointSpacing()
    return waypointSpacing
end

local function getWaypointsInScenario()
    return wpsInScenario
end

----------------------------------------------------------------------------------------------------
-- Interface
----------------------------------------------------------------------------------------------------

M.createWaypointsFromNodes = createWaypointsFromNodes
M.findWaypointsInLevel = findWaypointsInLevel
M.getNextWaypointData = getNextWaypointData
M.resetWaypoints = resetWaypoints
M.getLastWaypointData = getLastWaypointData
M.getWaypointData = getWaypointData
M.setLastPassedWaypoint = setLastPassedWaypoint
M.getWaypointSpacingValue = getWaypointSpacingValue
M.getLastPassedWaypoint = getLastPassedWaypoint
M.getLengthOfCurrentRoute = getLengthOfCurrentRoute
M.getWaypointSpacing = getWaypointSpacing
M.getWaypointsInScenario = getWaypointsInScenario
return M